package cc.smtweb.system.bpm.core.ui.builder;

import cc.smtweb.framework.core.common.SwMap;
import cc.smtweb.framework.core.db.sqlbuilder.AbstractSelectSqlBuilder;
import cc.smtweb.framework.core.db.sqlbuilder.SqlBuilder;
import cc.smtweb.framework.core.util.JsonUtil;
import cc.smtweb.system.bpm.core.exception.BpmException;
import cc.smtweb.system.bpm.core.ui.BpmPageContext;
import cc.smtweb.system.bpm.core.ui.IParamConst;
import cc.smtweb.system.bpm.core.ui.entity.dataset.BpmDataset;
import cc.smtweb.system.bpm.core.ui.entity.dataset.BpmField;
import cc.smtweb.system.bpm.engine.ui.entity.consts.UiEnum;
import cc.smtweb.system.bpm.engine.ui.entity.po.DynFormLogPO;
import lombok.Getter;
import org.apache.commons.lang3.StringUtils;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;

/**
 * 数据模型处理，读取和排序处理
 */
@Getter
public class TreeDataBuilder {
  private BpmField idField = null;
  private BpmField parentField = null;
  private BpmField orderField = null;
  private BpmField titleField = null;
  private BpmDataset dataset;
  private Long parentId;
  private Long[] orders;
  private String idName;

  public static TreeDataBuilder create(BpmPageContext ctx) {
    SwMap body = ctx.getBody();
    Long parentId = body.readLong(IParamConst.PARAM_PARENT_ID);
    Long[] orders = body.readLongArray(IParamConst.PARAM_ORDERS);

    if (orders == null) {
      throw new BpmException(IParamConst.PARAM_ORDERS + " 参数必填");
    }

    TreeDataBuilder treeDataBuilder = new TreeDataBuilder(ctx.getDataset(), parentId, orders);

    if (treeDataBuilder.getParentField() == null) {
      throw new BpmException("dataset未设置‘上级ID’字段");
    }

    if (treeDataBuilder.getOrderField() == null) {
      throw new BpmException("dataset未设置‘排序’字段");
    }
    return treeDataBuilder;
  }

  public TreeDataBuilder(BpmDataset dataset, Long parentId, Long[] orders) {
    this.dataset = dataset;
    this.parentId = parentId;
    this.orders = orders;
    for (Map.Entry<String, BpmField> entry: dataset.getFields().entrySet()) {
      BpmField bpmField = entry.getValue();

      switch (bpmField.getRefType()) {
        case UiEnum.FieldRefType.ID:
          idField = bpmField;
          break;
        case UiEnum.FieldRefType.PARENT_ID:
          parentField = bpmField;
          break;
        case UiEnum.FieldRefType.ORDER:
          orderField = bpmField;
          break;
        case UiEnum.FieldRefType.TITLE:
          titleField = bpmField;
          break;
        default:
          break;
      }
    }

    idName = idField != null ? idField.getName() : IParamConst.PARAM_ID;
  }

  public AbstractSelectSqlBuilder tableSqlBuilder() {
    AbstractSelectSqlBuilder sqlBuilder = SqlBuilder.createSelect(dataset.getDatabase(), dataset.getTable());

    if (idField != null) {
      sqlBuilder.add(idField.getFieldName() + " " + idField.getName());
    } else {
      sqlBuilder.add(dataset.readIdFieldName() + " id");
    }
    sqlBuilder.add(parentField.getFieldName() + " " + parentField.getName());
    sqlBuilder.add(titleField.getFieldName() + " " + titleField.getName());

    if (orderField != null) {
      sqlBuilder.addOrderBy(orderField.getFieldName() + "," + dataset.readIdFieldName());
    } else {
      sqlBuilder.addOrderBy(dataset.readIdFieldName());
    }

    return sqlBuilder;
  }

  public SwMap mapTreeFields(BpmDataset dataset, DynFormLogPO po) {
    if (StringUtils.isNotBlank(po.getDflContent())) {
      SwMap item = new SwMap();
      SwMap entity = JsonUtil.parse(po.getDflContent(), SwMap.class);

      item.put(parentField.getName(), entity.get(parentField.getFieldName()));
      item.put(titleField.getName(), entity.get(titleField.getFieldName()));
      item.put("id", po.getDflId());

      return item;
    }

    return null;
  }

  public SwMap updateOrder(DynFormLogPO po, Long parentId, int order) {
    SwMap entity;
    if (StringUtils.isNotBlank(po.getDflContent())) {
      entity = JsonUtil.parse(po.getDflContent(), SwMap.class);
    } else {
      entity = new SwMap();
    }

    entity.put(this.parentField.getFieldName(), parentId);
    entity.put(this.orderField.getFieldName(), order);

    return entity;
  }

  public static boolean longEquals(Long l1, Long l2) {
    if (l1 != null) {
      return l1.equals(l2);
    }

    return l2 == null;
  }

  private void addChild(SwMap parent, SwMap item) {
    List<SwMap> children = readChildren(parent);

    if (children == null) {
      children = new ArrayList<>();
      parent.put(IParamConst.TREE_CHILDREN, children);
    }

    children.add(item);
  }

  private boolean findAndAdd(SwMap parent, SwMap item) {
    if (longEquals(parent.readLong(idName), item.readLong(parentField.getName()))) {
      addChild(parent, item);
      return true;
    } else {
      List<SwMap> children = readChildren(parent);

      if (children != null) {
        for (SwMap child : children) {
          if (findAndAdd(child, item)) {
            return true;
          }
        }
      }
    }

    return false;
  }

  private List<SwMap> readChildren(SwMap parent) {
    return (List<SwMap>) parent.get(IParamConst.TREE_CHILDREN);
  }

  // 读取树结构
  public List<SwMap> buildTree(List<SwMap> list, boolean forceAdd) {
    SwMap root = new SwMap();

    if (list != null) {
      for (SwMap item: list) {
        if (!findAndAdd(root, item)) {
          for (SwMap it: list) {
            // TODO IParamConst.PARAM_ID => idField.getId()
            if (longEquals(it.readLong(idName), item.readLong(parentField.getName()))) {
              addChild(it, item);
              item = null;
              break;
            }
          }

          // not find item parent
          if (item != null && forceAdd) {
            addChild(root, item);
          }
        }
      }
    }

    List<SwMap> result = readChildren(root);

    if (result == null) {
      result = new ArrayList<>();
    }

    return result;
  }
}
