package cc.smtweb.system.bpm.web.design.form;

import cc.smtweb.framework.core.common.SwException;
import cc.smtweb.framework.core.common.SwMap;
import cc.smtweb.framework.core.cache.AbstractCache;
import cc.smtweb.framework.core.common.SwEnum;
import cc.smtweb.framework.core.db.cache.ModelTableCache;
import cc.smtweb.framework.core.db.vo.ModelField;
import cc.smtweb.framework.core.db.vo.ModelTable;
import cc.smtweb.framework.core.mvc.variable.SwVariableFactory;
import cc.smtweb.framework.core.session.UserSession;
import cc.smtweb.framework.core.util.CommUtil;
import cc.smtweb.framework.core.util.JsonUtil;
import cc.smtweb.framework.core.util.MapUtil;
import cc.smtweb.system.bpm.web.design.form.define.*;
import com.fasterxml.jackson.core.JsonProcessingException;
import org.apache.commons.lang3.StringUtils;

import java.util.*;

/**
 * Created by Akmm at 2022/4/20 18:45
 * 辅助类
 */
public class ModelFormHelper {
    /**
     * 从缓存获取Form对象
     *
     * @param formId
     * @return
     */
    public static ModelForm getFromCache(long formId) {
        AbstractCache<ModelForm> cache = ModelFormCache.getInstance();
        if (cache == null) return null;
        return cache.get(formId);
    }

    public static PageDatasets parsePageDataset(long formId) {
        ModelForm form = getFromCache(formId);
        if (form == null) return null;
        return parsePageDataset(form.getDataset());
    }

    public static PageDatasets parsePageDataset(String jsonStr) {
        if (StringUtils.isEmpty(jsonStr)) return null;
        PageDataset[] list = JsonUtil.parse(jsonStr, PageDataset[].class);
        return new PageDatasets(list);
    }

    /**
     * 根据json字符串解析对象
     *
     * @param jsonStr
     * @return
     */
    public static PageModel parsePageInfo(String jsonStr) {
        if (StringUtils.isEmpty(jsonStr)) return null;
        return JsonUtil.parse(jsonStr, PageModel.class);
    }

    public static SwMap parseFormOption(String jsonStr) {
        if (StringUtils.isEmpty(jsonStr)) return null;
        return JsonUtil.parse(jsonStr, SwMap.class);
    }

    /**
     * 页面传回的content，加工处理成待保存的json串
     * 主要处理内容：
     * 1、model的一些maxlen、默认值等信息，如果和数据库一致，则去掉
     * 2、db中的只保留表、字段和别名
     *
     * @param form
     * @return
     */
    public static String buildSaveModel(ModelForm form) {
        PageDatasets datasets = parsePageDataset(form.getDataset());
        PageModel pageInfo = parsePageInfo(form.getContent());
        if (pageInfo == null) return "";
        for (Map<String, Object> model : pageInfo.model) {
            buildSaveModelFields(datasets, pageInfo, model, true);
            buildSaveModelFields(datasets, pageInfo, model, false);
        }
        return JsonUtil.encodeString(pageInfo);
    }

    //处理model的fields和filters
    private static void buildSaveModelFields(PageDatasets datasets, PageModel pageInfo, Map<String, Object> model, boolean isField) {
        String db = (String) model.get("dataset");
        //没有配置db，配置有误，不处理
        if (StringUtils.isEmpty(db)) throw new SwException("model未配置数据集db，无法解析！");
        PageDataset pds = datasets.findById(db);
        String key = isField ? "fields" : "filters";
        List<Map<String, Object>> fields = (List<Map<String, Object>>) model.get(key);
        if (fields == null || fields.isEmpty()) return;
        for (Map<String, Object> field : fields) {
            String fn = (String) field.get("field");
            if (StringUtils.isEmpty(fn)) throw new SwException("model[" + db + "]." + key + "未配置字段名field，无法解析！");
            PageDatasetField pdf = isField ? pds.getField(fn) : pds.getFilter(fn);
            if (pdf == null) throw new SwException("model[" + db + "]." + key + "未找到定义的数据集字段(" + fn + ")，无法解析！");

            if (CommUtil.isStrEquals(MapUtil.readString(field, "label"), pdf.label)) {
                field.remove("label");
            }

            //必填字段
            if (pdf.isFieldNotNull() == MapUtil.readBool(field, "required")) {
                field.remove("required");
            }

            if (MapUtil.readInt(field, "maxlength") == SwEnum.DataType.instance.getByValue(pdf.dataType).dataLength) {
                field.remove("maxlength");
            }
        }
    }

    /**
     * 保存的数据集，删除掉一些不必要的字段信息
     * @param jsonStr
     * @return
     */
    public static String buildSaveDataset(String jsonStr) {
        PageDatasets datasets = parsePageDataset(jsonStr);
        if (datasets == null || datasets.list == null) return "";
        for (PageDataset dataSet: datasets.list) {
            buildSaveDataSetFields(dataSet.fields);
            buildSaveDataSetFields(dataSet.filters);
        }
        return JsonUtil.encodeString(datasets.list);
    }

    /**
     * 构建待保存的数据集
     *
     * @return
     * @throws JsonProcessingException
     */
    public static void buildSaveDataSetFields(List<? extends PageDatasetField> fields) {
        ModelTable table = null;
        for (PageDatasetField field: fields) {
            if (field.table <= 0) continue;
            if (table == null || table.getEntityId() != field.table) {
                table = ModelTableCache.getInstance().get(field.table);
            }
            if (table == null) continue;
            ModelField tf = table.findFieldByName(field.field);
            if (tf == null) continue;
            field.table_text = null;
            if (CommUtil.isStrEquals(tf.getTitle(), field.label)) field.label = null;
            if (CommUtil.isStrEquals(tf.getDataType(),field.dataType)) field.dataType = null;
            if (CommUtil.isStrEquals(tf.getRemark(), field.remark)) field.remark = null;
            if (CommUtil.isStrEquals(tf.getEditor(), field.editor)) field.editor = null;
        }
    }

    /**
     * 构建请求的数据集，补充表定义信息
     *
     * @return
     * @throws JsonProcessingException
     */
    public static String buildReqDataset(String jsonStr) {
        PageDatasets datasets = parsePageDataset(jsonStr);
        if (datasets == null || datasets.list == null) return "";
        return JsonUtil.encodeString(datasets.list);
    }


    /**
     * 根据数据库存储的信息，加工处理成页面要的，与buildsave逆向
     *
     * @param form
     * @return
     */
    public static String buildReqModel(ModelForm form) {
        PageDatasets datasets = parsePageDataset(form.getDataset());
        PageModel pageInfo = parsePageInfo(form.getContent());
        if (pageInfo == null) return "";
        for (Map<String, Object> model : pageInfo.model) {
            String db = (String) model.get("dataset");
            //没有配置db，配置有误，不处理
            if (StringUtils.isEmpty(db)) continue;
            PageDataset dataSet = datasets.findById(db);
            if (dataSet == null) continue;
            buildReqModelFields(dataSet, model, true);
            buildReqModelFields(dataSet, model, false);
        }
        return JsonUtil.encodeString(pageInfo);
    }

    //处理model的fields和filters
    private static void buildReqModelFields(PageDataset dataSet, Map<String, Object> model, boolean isField) {
        String key = isField ? "fields" : "filters";
        List<Map<String, Object>> fields = (List<Map<String, Object>>) model.get(key);
        if (fields == null || fields.isEmpty()) return;
        for (Map<String, Object> field : fields) {
            String fn = (String) field.get("field");
            if (StringUtils.isEmpty(fn)) continue;
            PageDatasetField pdf = isField ? dataSet.getField(fn) : dataSet.getFilter(fn);
            if (pdf == null) continue;
            if (!field.containsKey("label")) {
                field.put("label", pdf.label);
            }
            //必填字段
            if (!field.containsKey("required")) {
                field.put("required", pdf.isFieldNotNull());
            }

            if (!field.containsKey("maxlength")) {
                field.put("maxlength", SwEnum.DataType.instance.getByValue(pdf.dataType).dataLength);
            }
        }
    }

    /**
     * 根据数据库存储的信息，加工处理成页面要的，与buildsave逆向
     *
     * @param form
     * @return
     */
    public static String buildEngineModel(ModelForm form, SwMap params, UserSession us) {
        PageDatasets datasets = parsePageDataset(form.getDataset());
        if (datasets == null || datasets.list == null) return "";

        SwMap ret = new SwMap();
        PageModel pageInfo = parsePageInfo(form.getContent());
        if (pageInfo == null) return "";
        //form不用管
        if (pageInfo.form != null) ret.put("form", pageInfo.form);
        List<SwMap> listModel = new ArrayList<>();
        ret.put("model", listModel);
        //构建Model
        for (Map<String, Object> model : pageInfo.model) {
            String db = (String) model.get("dataset");
            PageDataset dataSet = datasets.findById(db);
            if (dataSet == null) throw new SwException("未找到指定的数据集定义！" + db);

            listModel.add(buildEngineModelMap(dataSet, model));
        }
        ret.put("extra", buildEngineExtra(pageInfo.option, params, us));
        return JsonUtil.encodeString(ret);
    }

    //构建model部分
    private static SwMap buildEngineModelMap(PageDataset dataSet, Map<String, Object> model) {
        SwMap ret = new SwMap();

        ret.put("id", dataSet.id);
        ret.put("name", dataSet.name);
        ret.put("label", dataSet.label);
        ret.put("type", dataSet.type);
        ret.put("idField", dataSet.idField);
        ret.put("lazy", dataSet.lazy);
        ret.put("canEdit", dataSet.canEdit);

        ret.put("fields", model.get("fields"));
        ret.put("filters", model.get("filters"));

        buildReqModelFields(dataSet, model, true);
        buildReqModelFields(dataSet, model, false);

        return ret;
    }

    //根据option构建Extra部分
    private static SwMap buildEngineExtra(Map<String, Object> option, SwMap params, UserSession us) {
        SwMap map = new SwMap();
        if (option == null) return map;
        //控件信息
        SwMap widget = new SwMap();
        map.put("widget", widget);

        List widgetIds = (List) option.get("widgetRef");
        for (Object v : widgetIds) {
            long id;
            if (v instanceof Integer) {
                id = (Integer)v;
            } else if (v instanceof Long) {
                id = (Long)v;
            } else throw new SwException("不能识别的控件Id：" + v);
            ModelForm widgetForm = getFromCache(id);
            if (widgetForm == null) throw new SwException("没有找到指定的控件定义！id=" + id);
            SwMap w = new SwMap();
            widget.put("w" + id, w);
            SwMap widgetOpts = parseFormOption(widgetForm.getOption());
            w.put("service", widgetForm.getService());
            if (widgetOpts != null) w.putAll(widgetOpts);
            w.put("define", widgetForm.getContent());
        }
        //构建变量
        SwMap mapVar = new SwMap();
        map.put("var", mapVar);
        List<String> vars = (List<String>) option.get("vars");
        for (String var : vars) {
            mapVar.put(var, SwVariableFactory.getInstance().calcVar(var, params, us));
        }
        return map;
    }

    //获取控件的filter信息
    public static List<SwMap> buildWidgetFilter(ModelForm bean) {
        List<SwMap> listRet = new ArrayList<>();
        PageDatasets datasets = ModelFormHelper.parsePageDataset(bean.getDataset());
        if (datasets == null || datasets.list == null) return listRet;
        for (PageDataset ds: datasets.list) {
            for (PageDatasetFilter filter: ds.filters) {
                if (SwEnum.FilterType.INPUT.value.equals(filter.type) || SwEnum.FilterType.PARAM.value.equals(filter.type)) {
                    SwMap fret = new SwMap();
                    fret.put("name", filter.name);
                    fret.put("label", filter.label);
                    listRet.add(fret);
                }
            }
        }
        listRet.sort((o1, o2) -> CommUtil.chineseCompare(o1.readString("label"), o2.readString("label")));
        return listRet;
    }
}
